#!/usr/bin/env python
# coding: utf-8

# ### 豆瓣评分的预测
# 
# 在这个项目中，我们要预测一部电影的评分，这个问题实际上就是一个分类问题。给定的输入为一段文本，输出为具体的评分。 在这个项目中，我们需要做：
# - 文本的预处理，如停用词的过滤，低频词的过滤，特殊符号的过滤等
# - 文本转化成向量，将使用三种方式，分别为tf-idf, word2vec以及BERT向量。 
# - 训练逻辑回归和朴素贝叶斯模型，并做交叉验证
# - 评估模型的准确率
# 
# 在具体标记为``TODO``的部分填写相应的代码。 

# In[18]:


#导入数据处理的基础包
import numpy as np
import pandas as pd

#导入用于计数的包
from collections import Counter

#导入tf-idf相关的包
from sklearn.feature_extraction.text import TfidfTransformer    
from sklearn.feature_extraction.text import CountVectorizer

#导入模型评估的包
from sklearn import metrics

#导入与word2vec相关的包
from gensim.models import KeyedVectors

#导入与bert embedding相关的包，关于mxnet包下载的注意事项参考实验手册
from bert_embedding import BertEmbedding
import mxnet

#包tqdm是用来对可迭代对象执行时生成一个进度条用以监视程序运行过程
from tqdm import tqdm

#导入其他一些功能包
import requests
import os,sys,time,copy
from collections import namedtuple, Counter, defaultdict
from sklearn.model_selection import train_test_split
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfVectorizer
from scipy.sparse import csr_matrix
import pickle
#========================================
#_python_dir_base='../../NPyBs'
#sys.path.append(_python_dir_base)
import NLBs,NLStrFilter,cfg

a=0
# ### 1. 读取数据并做文本的处理
# 你需要完成以下几步操作：
# - 去掉无用的字符如！&，可自行定义
# - 中文分词
# - 去掉低频词

# In[19]:


#读取数据
data_o = pd.read_csv('data/DMSC.csv')
#观察数据格式
data_o.head()


# In[20]:


#输出数据的一些相关信息
data_o.info()


# In[21]:


#只保留数据中我们需要的两列：Comment列和Star列
data_o = data_o[['Comment','Star']]
#观察新的数据的格式
data_o.head()


# In[22]:


# 这里的star代表具体的评分。但在这个项目中，我们要预测的是正面还是负面。我们把评分为1和2的看作是负面，把评分为3，4，5的作为正面
data_o['Star']=(data_o.Star/3).astype(int)
print(data_o.head(50))

#!!!!!total 21.56w rcds;
if cfg.rcd_num_to_proc>0:
	data=pd.DataFrame(data_o[:cfg.rcd_num_to_proc])
else:
	data=data_o

# #### 任务1： 去掉一些无用的字符

# In[ ]:

#====================== TODO1: 去掉一些无用的字符，自行定一个字符几何，并从文本中去掉
#    your to do 
a=0
flt=NLStrFilter.FilterCn()
flt.erase_space=False
flt.replace_symbo_space_as_comma=True	
#ret=flt.proc(inputSrc)
print('-----filter string ----------')
for i in range(len(data['Comment'])):
	a=data['Comment'][i]
	#b=NLStrFilter.filter_cn(a,erase_space=False,replace_symbo_space_as_comma=True)
	b=flt.proc(a)
	if np.random.rand()<0.005:
		print('^^^^filter proc^^^^ [%d]'%(i))
		print(a)
		print(b)
		print(data['Star'][i])
	#data['Comment'][i]=b #cannot use this
	data.iloc[i,data.columns.get_loc('Comment')]=b
a=0



# #### 任务2：使用分词对文本做分词

# In[12]:
'''
with open('./input/nlp_test0.txt',encoding='utf-8') as f:    # 读入文件时用GBK解码
    document = f.read()
    print(document)
    document_cut = jieba.cut(document)

    # 使用 suggest_freq(segment, tune=True) 可调节单个词语的词频，使其（或不能）被分出来
    jieba.suggest_freq('沙瑞金', True)
    jieba.suggest_freq('易学习', True)
    jieba.suggest_freq('王大路', True)
    jieba.suggest_freq('京州', True)


    result = ' '.join(document_cut)
    print(result)
    with open('./input/nlp_test2.txt', 'w', encoding='utf-8') as f2:   # 存储时用utf-8编码
        f2.write(result)
'''

# TODO2: 导入中文分词包jieba, 并用jieba对原始文本做分词
import jieba
a=0
def comment_cut(content):
	# TODO: 使用结巴完成对每一个comment的分词
	document_cut = jieba.cut(content)
	result = ' '.join(document_cut)
	result=result.replace(',','')#预处理之后，原文的标点和字符都转换为','；i.e.:只有',';

	return result

# 输出进度条
print('-----word cut ----------')
tqdm.pandas(desc='apply')
data['comment_processed'] = data['Comment'].progress_apply(comment_cut)


# In[13]:


# 观察新的数据的格式
data.head()


# #### 任务3：设定停用词并去掉停用词

# In[14]:


# TODO3: 设定停用词并从文本中去掉停用词

# 下载中文停用词表至data/stopWord.json中，下载地址:https://github.com/goto456/stopwords/
if not os.path.exists('data/stopWord.json'):
	stopWord = requests.get("https://raw.githubusercontent.com/goto456/stopwords/master/cn_stopwords.txt")
	with open("data/stopWord.json", "wb") as f:
		f.write(stopWord.content)

# 读取下载的停用词表，并保存在列表中
with open("data/stopWord.json","r",encoding='utf-8') as f:
	stopWords = f.read().split("\n")  
a=0

all_words='' #global variables

# 去除停用词;
def rm_stop_word(wordList):
	# your code, remove stop words
	# TODO
	global all_words   #use global;
	lst=wordList.split(' ')#
	outstr = ''
	for word in lst:
		if word not in stopWords:
			if word != '\t':
				outstr += word
				outstr += " "
		else:
			#print(word)
			pass
	all_words+=outstr
	return outstr	
print('-----stop words ----------')
#这行代码中.progress_apply()函数的作用等同于.apply()函数的作用，只是写成.progress_apply()函数才能被tqdm包监控从而输出进度条。
data['comment_processed'] = data['comment_processed'].progress_apply(rm_stop_word)

a=0


# In[ ]:


# 观察新的数据的格式
data.head()


# #### 任务4：去掉低频词，出现次数少于10次的词去掉

# In[ ]:

print('-----low-freq words ----------')
# TODO4: 去除低频词, 去掉词频小于10的单词，并把结果存放在data['comment_processed']里
t_low_freq_num=cfg.t_low_freq_num #<=3: low_freq; erase them;
all_words=all_words.split(' ')
words_counter=Counter(all_words)
words_counts = [x for x in words_counter.most_common(100)]

counter_wd_lst=list(words_counter.keys())
wd_array=np.array(list(words_counter.values()))
low_freq_ix=np.where(wd_array<=t_low_freq_num)[0]
#a1=[ all_words[i] if not i in low_freq_ix else '' for i in range(len(all_words))]
low_freq_wd_list=[ counter_wd_lst[i]  for i in range(len(all_words)) if i in low_freq_ix ] #filter

print('low_freq_thresh, low_freq_wd_num: %d, %d'%(t_low_freq_num,len(low_freq_wd_list)))
print(low_freq_wd_list[:100])

def rm_low_freq_word(wordList):
	# your code, remove stop words
	# TODO
	#global all_words   #use global;
	lst=wordList.split(' ')#
	outstr = ''
	for word in lst:
		if word not in low_freq_wd_list:
			if word != '\t':
				outstr += word
				outstr += " "
		else:
			#print(word)
			pass
	#all_words+=outstr
	return outstr	


data['comment_processed'] = data['comment_processed'].progress_apply(rm_low_freq_word)
print(data['Comment'].head(10))
print(data['comment_processed'].head(10))
print(data['Star'].head(10))
a=0

# ### 2. 把文本分为训练集和测试集
# 选择语料库中的20%作为测试数据，剩下的作为训练数据

# In[ ]:


# TODO5: 把数据分为训练集和测试集. comments_train（list)保存用于训练的文本，comments_test(list)保存用于测试的文本。 y_train, y_test是对应的标签（0、1）

#删除comment_processed为empty-str的记录;
x=np.array(list(data['comment_processed']))
y=np.array(list(data['Star']))
ix=np.where(x=='')
x=np.delete(x,ix,0)#np.delete(a,(1,2),0)#删除a的第二，三行。 
y=np.delete(y,ix,0)

test_ratio = 0.2

X_train, X_test, y_train, y_test = train_test_split(x, y, test_size=test_ratio, random_state=0)


comments_train, comments_test = X_train, X_test
y_train, y_test = y_train, y_test

print(data.shape)
print('==================sent2vec======================')
print('---------tfidf--------------')

# ### 3. 把文本转换成向量的形式
# 
# 在这个部分我们会采用三种不同的方式:
# - 使用tf-idf向量
# - 使用word2vec
# - 使用bert向量
# 
# 转换成向量之后，我们接着做模型的训练

# #### 任务6：把文本转换成tf-idf向量

# In[15]:


# TODO6: 把训练文本和测试文本转换成tf-idf向量。使用sklearn的feature_extraction.text.TfidfTransformer模块
#    请留意fit_transform和transform之间的区别。 常见的错误是在训练集和测试集上都使用 fit_transform，需要避免！ 
#    另外，可以留意一下结果是否为稀疏矩阵
print(' ~~~~ train-set~~~')
tfidf_vecc = TfidfVectorizer(token_pattern=r"(?u)\b\w+\b")#!!默认值只匹配长度≥2的单词
tfidf_train = tfidf_vecc.fit_transform(X_train)
tfidf_train = tfidf_train.toarray()
print(' ~~~~ test-set~~~')
tfidf_test=tfidf_vecc.transform(X_test)
tfidf_test=tfidf_test.toarray()
print (tfidf_train.shape, tfidf_test.shape)#train和test should same feat-length；
# #### 任务7：把文本转换成word2vec向量

# In[ ]:

print('---------word2vec by pre-trained--------------')

# 由于训练出一个高效的word2vec词向量往往需要非常大的语料库与计算资源，所以我们通常不自己训练Wordvec词向量，而直接使用网上开源的已训练好的词向量。
# data/sgns.zhihu.word是从https://github.com/Embedding/Chinese-Word-Vectors下载到的预训练好的中文词向量文件
# 使用KeyedVectors.load_word2vec_format()函数加载预训练好的词向量文件

model = KeyedVectors.load_word2vec_format('data/sgns.zhihu.word')


# In[ ]:


#预训练词向量使用举例
sample_wd_vec=model['今天']


# In[ ]:


# TODO7: 对于每个句子，生成句子的向量。具体的做法是：包含在句子中的所有单词的向量做平均。
vocabulary = model.vocab


def calc_word2vec_1doc(doc):
	tSum=np.zeros((len(sample_wd_vec)),dtype=np.float)#cannot (1,300)
	tNumWds=0
	for j,je in enumerate(doc.split(' ')):
		if not je in vocabulary:
			continue
		tVec=model[je]
		tSum=tSum+tVec #cannot +=;
		tNumWds+=1
	if tNumWds==0:
		a=0 #全部words都不在vocab中,若情况较多，则用sub-word进行处理;
		for j,je in enumerate(doc): #no .split()
			if not je in vocabulary:
				continue
			tVec=model[je]
			tSum=tSum+tVec #cannot +=;
			tNumWds+=1	
	if tNumWds==0:
		raise Exception('word2vec error1345')
		a=0
	has_dt=False
	if tNumWds>0:
		tSum=tSum/tNumWds #平均值;
		has_dt=True
	return list(tSum),has_dt	


word2vec_train=[]
word2vec_test=[]
num_no_wd_vec=0
print(' ~~~~ train-set~~~')
for i,e in enumerate(X_train):
	sentence_vec,has_dt=calc_word2vec_1doc(e)
	if not has_dt:
		num_no_wd_vec+=1
	word2vec_train.append(sentence_vec)	
	a=0
a=0
print(' ~~~~ test-set~~~')
for i,e in enumerate(X_test):
	sentence_vec,has_dt=calc_word2vec_1doc(e)
	if not has_dt:
		num_no_wd_vec+=1
	word2vec_test.append(sentence_vec)	
	a=0
word2vec_train=np.array(word2vec_train)	
word2vec_test=np.array(word2vec_test)
print (word2vec_train.shape, word2vec_test.shape)
a=0

# #### 任务8：把文本转换成bert向量

# In[16]:


# 导入gpu版本的bert embedding预训练的模型。
# 若没有gpu，则ctx可使用其默认值cpu(0)。但使用cpu会使程序运行的时间变得非常慢
# 若之前没有下载过bert embedding预训练的模型，执行此句时会花费一些时间来下载预训练的模型
#
#~~~~~~~~~~~~~~<解决“mxnet.base.MXNetError:GPU is not enabled”问题>>
#https://blog.csdn.net/SpadgerZ/article/details/103190481
#windows中只有：pip install mxnet-cu100
#
#(py3) C:\Users\Administrator>python
#Python 3.6.5 |Anaconda, Inc.| (default, Mar 29 2018, 13:32:41) [MSC v.1900 64 bit (AMD64)] on win32
#Type "help", "copyright", "credits" or "license" for more information.
#>>> import mxnet
#>>> a=mxnet.nd.array([1,2],mxnet.gpu())
#>>> a
#
#[1. 2.]
#<NDArray 2 @gpu(0)>
#
print('---------bert_vec--------------')
ctx = mxnet.gpu()
embedding = BertEmbedding(ctx=ctx)

# TODO8: 跟word2vec一样，计算出训练文本和测试文本的向量，仍然采用单词向量的平均。
sample_bert_wd_vec=embedding.embedding(['hello'])
bert_vec_len=len(sample_bert_wd_vec[0][1][0])

vocabulary=[]

bert_train=[]
bert_test=[]

def calc_word2vec_1doc(doc):
	bert_vec=embedding.embedding([doc])
	tSum=bert_vec[0][1][0]
	has_dt=True
	return list(tSum),has_dt	

print(' ~~~~ train-set~~~')
num_no_wd_vec=0
for i,e in enumerate(X_train):
	sentence_vec,has_dt=calc_word2vec_1doc(e)
	if not has_dt:
		num_no_wd_vec+=1
	bert_train.append(sentence_vec)	
	a=0
a=0
print(' ~~~~ test-set~~~')
for i,e in enumerate(X_test):
	sentence_vec,has_dt=calc_word2vec_1doc(e)
	if not has_dt:
		num_no_wd_vec+=1
	bert_test.append(sentence_vec)	
	a=0
	
bert_train=np.array(bert_train)
bert_test=np.array(bert_test)

print (bert_train.shape, bert_test.shape)


# In[ ]:

print('---------word2vec Done.--------------')

print (tfidf_train.shape, tfidf_test.shape)
print (word2vec_train.shape, word2vec_test.shape)
print (bert_train.shape, bert_test.shape)

a=0
with open('procedure_data_orig.pkl', 'wb') as f:
	pickle.dump(X_train, f)
	pickle.dump(X_test, f)	
	pickle.dump(y_train, f)
	pickle.dump(y_test, f)	
	
with open('procedure_data_tfidf.pkl', 'wb') as f:
	pickle.dump(tfidf_train, f)
	pickle.dump(tfidf_test, f)	
	
with open('procedure_data_wd2vec.pkl', 'wb') as f:
	pickle.dump(word2vec_train, f)
	pickle.dump(word2vec_test, f)
	
with open('procedure_data_bert.pkl', 'wb') as f:
	pickle.dump(bert_train, f)
	pickle.dump(bert_test, f)
	

a=0

"""
# ### 4. 训练模型以及评估
# 对如上三种不同的向量表示法，分别训练逻辑回归模型，需要做：
# - 搭建模型
# - 训练模型（并做交叉验证）
# - 输出最好的结果

# In[ ]:


# 导入逻辑回归的包
from sklearn.linear_model import LogisticRegression


# #### 任务9：使用tf-idf，并结合逻辑回归训练模型

# In[ ]:


# TODO9: 使用tf-idf + 逻辑回归训练模型，需要用gridsearchCV做交叉验证，并选择最好的超参数



print('TF-IDF LR test accuracy %s' % metrics.accuracy_score(y_test, tf_idf_y_pred))
#逻辑回归模型在测试集上的F1_Score
print('TF-IDF LR test F1_score %s' % metrics.f1_score(y_test, tf_idf_y_pred,average="macro"))


# #### 任务10：使用word2vec，并结合逻辑回归训练模型

# In[ ]:


# TODO10: 使用word2vec + 逻辑回归训练模型，需要用gridsearchCV做交叉验证，并选择最好的超参数



print('Word2vec LR test accuracy %s' % metrics.accuracy_score(y_test, word2vec_y_pred))
#逻辑回归模型在测试集上的F1_Score
print('Word2vec LR test F1_score %s' % metrics.f1_score(y_test, word2vec_y_pred,average="macro"))


# #### 任务11：使用bert，并结合逻辑回归训练模型

# In[ ]:


# TODO11: 使用bert + 逻辑回归训练模型，需要用gridsearchCV做交叉验证，并选择最好的超参数



print('Bert LR test accuracy %s' % metrics.accuracy_score(y_test, bert_y_pred))
#逻辑回归模型在测试集上的F1_Score
print('Bert LR test F1_score %s' % metrics.f1_score(y_test, bert_y_pred,average="macro"))


"""

# #### 任务12：对于以上结果请做一下简单的总结，按照1，2，3，4提取几个关键点，包括：
# - 结果说明什么问题？
# - 接下来如何提高？

# 1.
# 2.
# 3.
# 4.
# 5.
# 6.

