#!/usr/bin/env python
# coding: utf-8

# ### 豆瓣评分的预测
# 
# 在这个项目中，我们要预测一部电影的评分，这个问题实际上就是一个分类问题。给定的输入为一段文本，输出为具体的评分。 在这个项目中，我们需要做：
# - 文本的预处理，如停用词的过滤，低频词的过滤，特殊符号的过滤等
# - 文本转化成向量，将使用三种方式，分别为tf-idf, word2vec以及BERT向量。 
# - 训练逻辑回归和朴素贝叶斯模型，并做交叉验证
# - 评估模型的准确率
# 
# 在具体标记为``TODO``的部分填写相应的代码。 

# In[18]:


#导入数据处理的基础包
import numpy as np
import pandas as pd

#导入用于计数的包
from collections import Counter

#导入tf-idf相关的包
from sklearn.feature_extraction.text import TfidfTransformer    
from sklearn.feature_extraction.text import CountVectorizer

#导入模型评估的包
from sklearn import metrics

#导入与word2vec相关的包
#from gensim.models import KeyedVectors

#导入与bert embedding相关的包，关于mxnet包下载的注意事项参考实验手册
#from bert_embedding import BertEmbedding
#import mxnet

#包tqdm是用来对可迭代对象执行时生成一个进度条用以监视程序运行过程
from tqdm import tqdm

#导入其他一些功能包
import requests
import os,sys,time,copy
from collections import namedtuple, Counter, defaultdict
from sklearn.model_selection import train_test_split
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfVectorizer

from sklearn.model_selection import GridSearchCV
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.naive_bayes import MultinomialNB 
from sklearn.neighbors import KNeighborsClassifier

# 评估
from sklearn import metrics
from sklearn.metrics import classification_report
from sklearn.model_selection import cross_val_score
from sklearn.metrics import roc_auc_score
from sklearn.metrics import roc_curve
from sklearn.metrics import confusion_matrix
from sklearn.metrics import accuracy_score

from scipy.sparse import csr_matrix
import pickle
#========================================
#_python_dir_base='../../NPyBs'
#sys.path.append(_python_dir_base)
import NLBs,NLStrFilter,cfg
#====================================================================
def logistic_reg_fit_test(x_train,y_train,x_test,y_test,feat_des):
	print('----- feature-des:%s ------'%(feat_des))
	solver_type='liblinear'
	print('solver_type:%s'%(solver_type))
	logreg_cv = LogisticRegression(solver=solver_type,random_state=0)
	cv_dict = {0: 'Logistic Regression'}
	cv_models=[logreg_cv]	
	
	# TODO9: 使用tf-idf + 逻辑回归训练模型，需要用gridsearchCV做交叉验证，并选择最好的超参数
	#只是设置fold,没有调参
	print('general fit without parms tuning')
	for i,model in enumerate(cv_models):
		print("{} Test Accuracy: {}".format(
			cv_dict[i],cross_val_score(model, x_train, y_train, cv=10, scoring ='accuracy').mean()))
	'''
	~~~~~~~~output of above
	Logistic Regression Test Accuracy: 0.8622282265846419	
	'''
	#调参
	print('params tuning by grid-search')
	param_grid = {'C': np.logspace(-4, 4, 5), #(-4, 4, 50),
				  'penalty':['l1', 'l2']}
	clf = GridSearchCV(LogisticRegression(
		random_state=0,solver=solver_type), param_grid,cv=10, verbose=0,n_jobs=-1)
	best_model = clf.fit(x_train,y_train)
	print(best_model.best_estimator_)
	print("The mean accuracy of the model is:",best_model.score(x_test,y_test))	
	'''
	#~~~~~~~~~~~output of above
	LogisticRegression(C=1.7575106248547894, class_weight=None, dual=False,
					   fit_intercept=True, intercept_scaling=1, l1_ratio=None,
					   max_iter=100, multi_class='warn', n_jobs=None, penalty='l2',
					   random_state=0, solver='warn', tol=0.0001, verbose=0,
					   warm_start=False)
	0.8637931034482759
	'''
	#利用上述的最优参数来建模和重新训练;
	print('use best params to fit new model')
	logreg = LogisticRegression(C=best_model.best_estimator_.C,
								penalty=best_model.best_estimator_.penalty, 
								random_state=0,solver=solver_type)
	logreg.fit(x_train, y_train)
	print('predict')
	y_pred = logreg.predict(x_test)	
	print('return')
	return y_pred

#====================================================================
if __name__ == "__main__":

	procedure_dt_dir='procedure_data_log-20k-docs/'
	'''
	with open('procedure_data_orig.pkl', 'wb') as f:
		pickle.dump(X_train, f)
		pickle.dump(X_test, f)	
		pickle.dump(y_train, f)
		pickle.dump(y_test, f)	

	with open('procedure_data_tfidf.pkl', 'wb') as f:
		pickle.dump(tfidf_train, f)
		pickle.dump(tfidf_test, f)	

	with open('procedure_data_wd2vec.pkl', 'wb') as f:
		pickle.dump(word2vec_train, f)
		pickle.dump(word2vec_test, f)

	with open('procedure_data_bert.pkl', 'wb') as f:
		pickle.dump(bert_train, f)
		pickle.dump(bert_test, f)

	'''

	with open(procedure_dt_dir+'procedure_data_orig.pkl', 'rb') as dt:
		X_train = pickle.load(dt)# 加载数据		
		X_test  = pickle.load(dt) 
		y_train = pickle.load(dt)
		y_test  = pickle.load(dt)

	with open(procedure_dt_dir+'procedure_data_tfidf.pkl', 'rb') as dt:
		tfidf_train= pickle.load(dt)# 加载数据	
		tfidf_test= pickle.load(dt)

	with open(procedure_dt_dir+'procedure_data_wd2vec.pkl', 'rb') as dt:
		word2vec_train = pickle.load(dt)# 加载数据	
		word2vec_test = pickle.load(dt)# 加载数据	

	with open(procedure_dt_dir+'procedure_data_bert.pkl', 'rb') as dt:
		bert_train = pickle.load(dt)# 加载数据	
		bert_test = pickle.load(dt)# 加载数据	

	print (tfidf_train.shape, tfidf_test.shape)
	print (word2vec_train.shape, word2vec_test.shape)
	print (bert_train.shape, bert_test.shape)
	

	# ### 4. 训练模型以及评估
	# 对如上三种不同的向量表示法，分别训练逻辑回归模型，需要做：
	# - 搭建模型
	# - 训练模型（并做交叉验证）
	# - 输出最好的结果
	
	# In[ ]:
	
	
	# 导入逻辑回归的包
	# #### 任务9：使用tf-idf，并结合逻辑回归训练模型
	tf_idf_y_pred=logistic_reg_fit_test(tfidf_train, y_train, tfidf_test, y_test, "tfidf")
	# In[ ]:
	
	print('TF-IDF LR test accuracy %s' % metrics.accuracy_score(y_test, tf_idf_y_pred))
	#逻辑回归模型在测试集上的F1_Score
	print('TF-IDF LR test F1_score %s' % metrics.f1_score(y_test, tf_idf_y_pred,average="macro"))



	# #### 任务10：使用word2vec，并结合逻辑回归训练模型
	
	# In[ ]:
	
	
	# TODO10: 使用word2vec + 逻辑回归训练模型，需要用gridsearchCV做交叉验证，并选择最好的超参数
	word2vec_y_pred=logistic_reg_fit_test(word2vec_train, y_train, word2vec_test, y_test, "word2vec")
	
	print('Word2vec LR test accuracy %s' % metrics.accuracy_score(y_test, word2vec_y_pred))
	#逻辑回归模型在测试集上的F1_Score
	print('Word2vec LR test F1_score %s' % metrics.f1_score(y_test, word2vec_y_pred,average="macro"))
	
	
	# #### 任务11：使用bert，并结合逻辑回归训练模型
	
	# In[ ]:
	
	
	# TODO11: 使用bert + 逻辑回归训练模型，需要用gridsearchCV做交叉验证，并选择最好的超参数
	bert_y_pred=logistic_reg_fit_test(bert_train, y_train, bert_test, y_test, "bert")
	
	
	print('Bert LR test accuracy %s' % metrics.accuracy_score(y_test, bert_y_pred))
	#逻辑回归模型在测试集上的F1_Score
	print('Bert LR test F1_score %s' % metrics.f1_score(y_test, bert_y_pred,average="macro"))
	
	
	"""
	
	# #### 任务12：对于以上结果请做一下简单的总结，按照1，2，3，4提取几个关键点，包括：
	# - 结果说明什么问题？
	# - 接下来如何提高？
	
	# 1.
	# 2.
	# 3.
	# 4.
	# 5.
	# 6.	
	"""
	
		
	print('==== exit ====')
	os._exit(0)
